#ifndef STREU_CPLUSLIB_CALL_H_INCLUDED
#define STREU_CPLUSLIB_CALL_H_INCLUDED

#include "cpluslib/smartptr.h"

#define DECLARE_CALLABLE(NAME, ARGS, CALL, ABSARGS)                          \
class NAME {                                                                 \
    class Handler {                                                          \
     public:                                                                 \
        virtual ~Handler() { }                                               \
        virtual RV call(ARGS) = 0;                                           \
        virtual Handler* clone() = 0;                                        \
    };                                                                       \
                                                                             \
    class Static : public Handler {                                          \
        RV (*func)(ARGS);                                                    \
     public:                                                                 \
        Static(RV (*func)(ARGS)) : func(func) { }                            \
        RV call(ARGS) { return func(CALL); }                                 \
        Static* clone() { return new Static(*this); }                        \
    };                                                                       \
                                                                             \
    template<class T>                                                        \
    class Object : public Handler {                                          \
        T* obj;                                                              \
        RV (T::*func)(ARGS);                                                 \
     public:                                                                 \
        Object(T* obj, RV (T::*func)(ARGS)) : obj(obj), func(func) { }       \
        RV call(ARGS) { return (obj->*func)(CALL); }                         \
        Object* clone() { return new Object(*this); }                        \
    };                                                                       \
                                                                             \
    template<class T>                                                        \
    class SmartObj : public Handler {                                        \
        Ptr<T> obj;                                                          \
        RV (T::*func)(ARGS);                                                 \
     public:                                                                 \
        SmartObj(Ptr<T> obj, RV (T::*func)(ARGS)) : obj(obj), func(func) { } \
        RV call(ARGS) { return ((*obj).*func)(CALL); }                       \
        SmartObj* clone() { return new SmartObj(*this); }                    \
    };                                                                       \
                                                                             \
    static void nullop(ABSARGS) { }                                          \
                                                                             \
    Handler* handler;                                                        \
 public:                                                                     \
    NAME()                                                                   \
        : handler(new Static(nullop)) { }                                    \
                                                                             \
    NAME(RV (*func)(ARGS))                                                   \
        : handler(new Static(func)) { }                                      \
                                                                             \
    template<class T>                                                        \
    NAME(T& obj, RV (T::*func)(ARGS))                                        \
        : handler(new Object<T>(&obj, func)) { }                             \
                                                                             \
    template<class T>                                                        \
    NAME(Ptr<T> obj, RV (T::*func)(ARGS))                                    \
        : handler(new SmartObj<T>(obj, func)) { }                            \
                                                                             \
    NAME(const NAME& rhs)                                                    \
        : handler(rhs.handler->clone()) { }                                  \
                                                                             \
    ~NAME() { delete handler; }                                              \
                                                                             \
    NAME& operator=(const NAME& rhs) {                                       \
        if (rhs.handler != handler) {                                        \
            Handler* n = rhs.handler->clone();                               \
            delete handler;                                                  \
            handler = n;                                                     \
        }                                                                    \
        return *this;                                                        \
    }                                                                        \
                                                                             \
    RV operator()(ARGS) const { return handler->call(CALL); }                \
}

#define EMPTY
#define AND ,

template<class RV>
DECLARE_CALLABLE(Callable0, EMPTY, EMPTY, EMPTY);
template<class RV, class A>
DECLARE_CALLABLE(Callable1, A a, a, A);
template<class RV, class A, class B>
DECLARE_CALLABLE(Callable2, A a AND B b, a AND b, A AND B);

#undef DECLARE_CALLABLE

#define DECLARE_MAKE(RET, TPLARGS, ARGS)                                \
template<class RV TPLARGS>                                              \
RET makeCallable(RV (*fun)(ARGS)) { return RET(fun); }                  \
template<class RV TPLARGS, class T>                                     \
RET makeCallable(T& o, RV (T::*fun)(ARGS)) { return RET(o, fun); }      \
template<class RV TPLARGS, class T>                                     \
RET makeCallable(Ptr<T> o, RV (T::*fun)(ARGS)) { return RET(o, fun); }

DECLARE_MAKE(Callable0<RV>,             EMPTY,                   EMPTY)
DECLARE_MAKE(Callable1<RV AND A>,       AND class A,             A)
DECLARE_MAKE(Callable2<RV AND A AND B>, AND class A AND class B, A AND B)

#undef DECLARE_MAKE
#undef AND
#undef EMPTY

#endif
